/*
** Copyright (C) 2004 Level Control Systems <jaf@lcsaudio.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
*/


#include <qapplication.h>
#include <qcheckbox.h>
#include <qcombobox.h>
#include <qgroupbox.h>
#include <qmessagebox.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qlistview.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qstatusbar.h>
#include <qtimer.h>
#include "audiomove/AudioMoveConstants.h"
#include "audiomove/AudioMoveWindow.h"
#include "audiomove/ErrorIOThread.h"
#include "audiomove/RawIOThread.h"
#include "audiomove/LibSndFileIOThread.h"
#include "audiomove/SampleRateThread.h"
#include "audiomove/MiscFunctions.h"
#include "util/StringTokenizer.h"

#if defined(WIN32)
# include "win32dirent.h"
#else
# include <dirent.h>
#endif

namespace audiomove {

enum {
   MOVE_COLUMN_JOB_ID,
   MOVE_COLUMN_SOURCE_FILE,
   MOVE_COLUMN_DEST_FILE,
   MOVE_COLUMN_PERCENT_DONE,
   MOVE_COLUMN_STATUS,
   MOVE_COLUMN_LENGTH,
   MOVE_COLUMN_CHANNELS,
   MOVE_COLUMN_SOURCE_FORMAT,
   MOVE_COLUMN_SOURCE_RATE,
   MOVE_COLUMN_SOURCE_WIDTH,
   MOVE_COLUMN_DEST_FORMAT,
   MOVE_COLUMN_DEST_RATE,
   MOVE_COLUMN_DEST_WIDTH,
   MOVE_COLUMN_QUALITY,
   NUM_MOVE_COLUMNS
};

enum {
   BUFFER_RETURNED_EVENT_TYPE = QEvent::User+600
};

class BufferReturnedEvent : public QCustomEvent
{
public:
   BufferReturnedEvent(MessageRef msg, bool isFromSetup) : QCustomEvent(BUFFER_RETURNED_EVENT_TYPE), _msg(msg), _isFromSetup(isFromSetup) {/* empty */}

   MessageRef GetMessage() const {return _msg;}

   bool IsFromSetup() const {return _isFromSetup;}

private:
   MessageRef _msg;
   bool _isFromSetup;
};

enum {
   SETUP_COMMAND_SETUP = 1000,
   SETUP_REPLY_RESULTS
};

#define SETUP_NAME_SETUPRESULT  "rslt"
#define SETUP_NAME_SOURCEFILE   "file"
#define SETUP_NAME_TARGETFORMAT "tfmt"
#define SETUP_NAME_TARGETRATE   "trat"
#define SETUP_NAME_TARGETWIDTH  "twid"
#define SETUP_NAME_DESTDIR      "ddir"
#define SETUP_NAME_INPUTTHREAD  "ithr"
#define SETUP_NAME_OUTPUTTHREAD "othr"
#define SETUP_NAME_QUALITY      "qual"
#define SETUP_NAME_ERROR        "errr"

void AdjustComboBoxWidth(QComboBox * combo)
{
#ifdef __APPLE__
   QFontMetrics fm = combo->fontMetrics();
   int maxStringWidth = 0;
   for (int j=combo->count()-1; j>=0; j--) maxStringWidth = muscleMax(maxStringWidth, fm.width(combo->text(j)));
   combo->setMaximumWidth(maxStringWidth+32);
   combo->setMinimumWidth(maxStringWidth+32);
#endif
   (void) combo;  // shut the compiler up
}

static bool ParseBool(const String & text, bool defVal)
{
   String s = text.ToUpperCase();
   if ((s.IndexOf('1') >= 0)||(s.IndexOf('Y') >= 0)||(s.IndexOf('T') >= 0)) return true;
   if ((s.IndexOf('0') >= 0)||(s.IndexOf('N') >= 0)||(s.IndexOf('F') >= 0)) return false;
   return defVal;
}

// Returns (dir + GetPathSepChar() + file) properly, even if dir and file already have GetPathSepChar() in them
static String AddPath(const String & dir, const String & file)
{
   String d = dir;
   while(d.EndsWith(GetPathSepChar())) d = d.Substring(0, d.Length()-GetPathSepChar().Length());

   String f = file;
   while(f.StartsWith(GetPathSepChar())) f = f.Substring(GetPathSepChar().Length());
  
   return d + GetPathSepChar() + f; 
}

/** This thread handles the initial opening and setup of file items, so that the GUI isn't blocked during setup */
class AudioSetupThread : public Thread
{
public:
   AudioSetupThread(AudioMoveWindow * win) : _window(win) {/* empty */
   }

protected:
   virtual status_t MessageReceivedFromOwner(MessageRef msgRef, uint32 numLeft)
   {
      Message * msg = msgRef();
      if ((msg)&&(msg->what == SETUP_COMMAND_SETUP))
      {
         String sourceFile, destDir;
         uint32 targetFormat, targetRate, targetWidth, quality;
         if ((msg->FindString(SETUP_NAME_SOURCEFILE, sourceFile)              == B_NO_ERROR)&&
             (msg->FindString(SETUP_NAME_DESTDIR,    destDir)                 == B_NO_ERROR)&&
             (msg->FindInt32(SETUP_NAME_TARGETFORMAT, (int32*) &targetFormat) == B_NO_ERROR)&&
             (msg->FindInt32(SETUP_NAME_TARGETRATE,   (int32*) &targetRate)   == B_NO_ERROR)&&
             (msg->FindInt32(SETUP_NAME_TARGETWIDTH,  (int32*) &targetWidth)  == B_NO_ERROR)&&
             (msg->FindInt32(SETUP_NAME_QUALITY,      (int32*) &quality)      == B_NO_ERROR)) DoSetup(sourceFile, destDir, targetFormat, targetRate, targetWidth, quality);
         return B_NO_ERROR;
      }
      else return Thread::MessageReceivedFromOwner(msgRef, numLeft);
   }

   void DoSetup(const String & sourceFile, const String & destDir, uint32 targetFormat, uint32 targetSampleRate, uint32 targetSampleWidth, uint32 quality)
   {
      MessageRef ret = GetMessageFromPool(SETUP_REPLY_RESULTS);
      if (ret())
      {
         DoSetupAux(*ret(), sourceFile, destDir, targetFormat, targetSampleRate, targetSampleWidth, quality);
         QApplication::postEvent(_window, new BufferReturnedEvent(ret, true));
      }
   }

   void DoSetupAux(Message & addMsgTo, const String & sourceFile, const String & destDir, uint32 targetFormat, uint32 targetSampleRate, uint32 targetSampleWidth, uint32 quality)
   {
      // First see if the sourceFile is really a directory.  If so, then we will recurse to its contents
      DIR * dir = opendir(sourceFile());
      if (dir)
      {
         String newDestDir = AddPath(destDir, (sourceFile.EndsWith("/")?sourceFile.Substring(0,sourceFile.Length()-1):sourceFile).Substring("/"));
         struct dirent * de;
         while((de = readdir(dir)) != NULL)
         {
            const char * n = de->d_name;
            if (n[0] != '.') DoSetupAux(addMsgTo, AddPath(sourceFile, n), newDestDir, targetFormat, targetSampleRate, targetSampleWidth, quality);
         }
         closedir(dir);
      }
      else
      {
         MessageRef subMsg = GetMessageFromPool();
         if (subMsg())
         {
            QString errorString;

            AudioMoveThreadRef inputThread;
            {
               // First, see if we can open the input file as an LCS raw file
               inputThread.SetRef(new RawIOThread(sourceFile), NULL);
               if (inputThread()->OpenFile() != B_NO_ERROR)
               {
                  // No?  Okay, then try opening it using LibSndFile
                  inputThread.SetRef(new LibSndFileIOThread(sourceFile), NULL);
                  if (inputThread()->OpenFile() != B_NO_ERROR)
                  {
                     // No?  Okay, then we're flummoxed; just report an error
                     inputThread.SetRef(new ErrorIOThread(sourceFile), NULL);
                     errorString = qApp->translate("AudioSetupThread", "Couldn't open input file");
                  }
               }
            }

            if (targetFormat == AUDIO_FORMAT_SOURCE) targetFormat = inputThread()->GetInputFileFormat();
            if (targetFormat == AUDIO_FORMAT_LCSRAW) 
            {
               targetSampleRate  = AUDIO_RATE_48000;  // LCS native supports only this rate!
               targetSampleWidth = AUDIO_WIDTH_FLOAT; // LCS native supports only this width!
            }
            else
            {
               if (targetSampleRate  == AUDIO_RATE_SOURCE)  targetSampleRate  = inputThread()->GetInputFileSampleRate();
               if (targetSampleWidth == AUDIO_WIDTH_SOURCE) targetSampleWidth = inputThread()->GetInputFileSampleWidth();
            }

            String dd = destDir;
            if (dd.EndsWith("/") == false) dd += '/';
            String baseFileName = sourceFile.Substring("/");
            String destFile = dd + baseFileName;

            const char * newExt = GetAudioFormatExtensions(targetFormat);
            if (newExt) destFile = ReplaceExtension(destFile, newExt);

            if (targetSampleRate == AUDIO_RATE_SOURCE)  targetSampleRate  = inputThread()->GetInputFileSampleRate();
            if (targetSampleRate == AUDIO_WIDTH_SOURCE) targetSampleWidth = inputThread()->GetInputFileSampleWidth();

            targetSampleRate = GetSampleRateValueFromCode(targetSampleRate);  // ensure that we have the actually value in Hz

            inputThread()->CloseFile();  // we'll open it again when we are ready to do the actual conversion

            AudioMoveThreadRef outputThread;
            if (errorString.length() > 0) outputThread.SetRef(new ErrorIOThread(destFile), NULL);
            else
            {
               switch(targetFormat)
               {
                  case AUDIO_FORMAT_WAV: case AUDIO_FORMAT_AIFF: 
                     outputThread.SetRef(new LibSndFileIOThread(destFile, targetFormat, targetSampleWidth, targetSampleRate, inputThread()->GetFileStreams()), NULL);
                  break;

                  case AUDIO_FORMAT_LCSRAW:                      
                     outputThread.SetRef(new RawIOThread(destFile, inputThread()->GetFileStreams()), NULL);
                  break;

                  default:  
                     errorString = qApp->translate("AudioSetupThread", "Unknown output format");
                     outputThread.SetRef(new ErrorIOThread(destFile), NULL);
                  break;
               }
            }

            if ((subMsg()->AddInt32(SETUP_NAME_QUALITY,    quality)                   == B_NO_ERROR)&&
                (subMsg()->AddTag(SETUP_NAME_INPUTTHREAD,  inputThread.GetGeneric())  == B_NO_ERROR)&&
                (subMsg()->AddTag(SETUP_NAME_OUTPUTTHREAD, outputThread.GetGeneric()) == B_NO_ERROR)&&
                ((errorString.length() == 0)||(subMsg()->AddString(SETUP_NAME_ERROR, FromQ(errorString)) == B_NO_ERROR))) (void) addMsgTo.AddMessage(SETUP_NAME_SETUPRESULT, subMsg);
         }
      }
   }

   const char * GetAudioFormatExtensions(uint32 format) const
   {
      switch(format)
      {
         case AUDIO_FORMAT_WAV:    return ".wav";
         case AUDIO_FORMAT_AIFF:   return ".aiff;.aif";  // two possibilities for this one
         case AUDIO_FORMAT_LCSRAW: return ".lcs";
      }
      return NULL;
   }

   String ReplaceExtension(const String & ss, const String & ext) const
   {
      String s = ss;
      bool useUpperCase = false;

      // First, remove any existing extension...
      for (uint32 i=0; i<NUM_AUDIO_FORMATS; i++)
      {
         const char * e = GetAudioFormatExtensions(i);
         if (e)
         {
            StringTokenizer tok(e, ";");
            const char * nextExt;
            while((nextExt = tok()) != NULL)
            {
               String es(nextExt);
               if (s.EndsWithIgnoreCase(es))
               {
                  uint32 extStart = s.Length()-es.Length();
                  useUpperCase = muscleInRange(s[extStart+1], 'A', 'Z');
                  s = s.Substring(0, extStart);
                  break;
               }
            }
         }
      }

      // Then add our new extension
      return s + (useUpperCase ? ext.ToUpperCase() : ext.ToLowerCase()).Substring(0, ";");
   }

   AudioMoveWindow * _window;
};

AudioMoveItem :: AudioMoveItem(uint32 tag, AudioMoveWindow * owner, QListView * lv, QListViewItem * after, AudioMoveThreadRef inputThread, AudioMoveThreadRef convertThread, AudioMoveThreadRef outputThread) : QListViewItem(lv, after), _tag(tag), _internalThreadsStarted(false), _status(MOVE_STATUS_WAITING), _samplesComplete(0), _totalSamples(inputThread()->GetFileFrames()*inputThread()->GetFileStreams()), _numActiveBuffers(0), _owner(owner), _lastUpdateTime(0)
{
   _samplesLeft                = _totalSamples;
   _threads[STAGE_SOURCE]      = inputThread;
   _threads[STAGE_CONVERSION]  = convertThread;
   _threads[STAGE_DESTINATION] = outputThread;

   for (uint32 i=0; i<ARRAYITEMS(_threads)-1; i++) _threads[i]()->SetDownstreamAcceptor(_threads[i+1]());
   _threads[ARRAYITEMS(_threads)-1]()->SetDownstreamAcceptor(owner);
}

AudioMoveItem :: ~AudioMoveItem()
{
   ShutdownInternalThreads();
}

void AudioMoveItem :: ShutdownInternalThreads()
{
   // Make sure our worker threads all get shut down safely
   for (uint32 i=0; i<ARRAYITEMS(_threads); i++) _threads[i]()->ShutdownInternalThread();
}

void AudioMoveItem :: SetStatus(uint32 s)
{
   if (s != _status)
   {
      _status = s;
      Update(true);  // update all cells so they all get the new background color
      _owner->UpdateButtons();
   }
}

status_t AudioMoveItem :: StartInternalThreads()
{
   if (_internalThreadsStarted == false)
   {
      MessageRef startup = GetMessageFromPool(_tag);
      if ((startup())&&(startup()->AddBool(AUDIOMOVE_NAME_OPENFILE, true) == B_NO_ERROR)&&(_threads[STAGE_SOURCE]()->MessageReceivedFromUpstream(startup) == B_NO_ERROR))
      {
         _numActiveBuffers++;  // the OPENFILE command counts as a buffer!
         for (uint32 i=0; i<ARRAYITEMS(_threads); i++) 
         {
            if (_threads[i]()->StartInternalThread() != B_NO_ERROR) 
            {
               SetStatus(MOVE_STATUS_ERROR);
               ShutdownInternalThreads();
               SetStatusString(qApp->translate("AudioMoveItem", "Couldn't spawn worker thread!"));
               Update(true);
               return B_ERROR;
            }
         }
         _internalThreadsStarted = true;
      }
      else return B_ERROR;
   }
   return B_NO_ERROR;
}

void AudioMoveItem :: SetStatusString(const QString & s)
{
   _statusString = s;
   Update(true);
}

void AudioMoveItem :: paintCell(QPainter * p, const QColorGroup & cg, int column, int width, int align)
{
   QListViewItem::paintCell(p, _owner->GetStatusColorGroup(_status, cg), column, width, align);
}

void AudioMoveItem :: Update(bool full)
{
   if (full)
   {
      const AudioMoveThread  * input  = _threads[STAGE_SOURCE]();
      const SampleRateThread * conv   = (SampleRateThread *) _threads[STAGE_CONVERSION]();
      const AudioMoveThread  * output = _threads[STAGE_DESTINATION]();

      setText(MOVE_COLUMN_JOB_ID,        ToQ("%1").arg(_tag));
      setText(MOVE_COLUMN_SOURCE_FILE,   ToQ(input->GetFileName()()));
      setText(MOVE_COLUMN_SOURCE_FORMAT, _owner->GetFileFormatName(input->GetInputFileFormat()));
      setText(MOVE_COLUMN_SOURCE_RATE,   _owner->GetSampleRateName(input->GetInputFileSampleRate(), true));
      setText(MOVE_COLUMN_SOURCE_WIDTH,  _owner->GetSampleWidthName(input->GetInputFileSampleWidth(), true));
      setText(MOVE_COLUMN_LENGTH,        _owner->GetFileDurationName(input->GetFileFrames(), input->GetInputFileSampleRate()));
      setText(MOVE_COLUMN_CHANNELS,      ToQ("%1").arg(input->GetFileStreams()));
      setText(MOVE_COLUMN_DEST_FILE,     ToQ(output->GetFileName()()));
      setText(MOVE_COLUMN_DEST_FORMAT,   _owner->GetFileFormatName(output->GetOutputFileFormat()));
      setText(MOVE_COLUMN_DEST_RATE,     _owner->GetSampleRateName(output->GetOutputFileSampleRate(), true));
      setText(MOVE_COLUMN_DEST_WIDTH,    _owner->GetSampleWidthName(output->GetOutputFileSampleWidth(), true));
      setText(MOVE_COLUMN_QUALITY,       _owner->GetConversionQualityName(conv->GetConversionQuality()));
      setText(MOVE_COLUMN_STATUS,        (_statusString.length() > 0) ? _statusString : _owner->GetStatusName(GetStatus()));
   }
   char buf[64]; sprintf(buf, "%.01f%%", GetPercentDone()*100.0f);
   setText(MOVE_COLUMN_PERCENT_DONE,  ToQ(buf));
}

int AudioMoveItem :: compare(QListViewItem * i, int col, bool ascending) const
{
   const AudioMoveItem * rhs = (const AudioMoveItem *)i;
   const AudioMoveThread * myInput   = _threads[STAGE_SOURCE]();
   const AudioMoveThread * myOutput  = _threads[STAGE_DESTINATION]();
   const SampleRateThread * myConv   = (SampleRateThread *) _threads[STAGE_CONVERSION]();
   const AudioMoveThread * hisInput  = rhs->_threads[STAGE_SOURCE]();
   const AudioMoveThread * hisOutput = rhs->_threads[STAGE_DESTINATION]();
   const SampleRateThread * hisConv  = (SampleRateThread *) rhs->_threads[STAGE_CONVERSION]();
   switch(col)
   {
      case MOVE_COLUMN_JOB_ID:        return muscleCompare(_tag, rhs->_tag);
      case MOVE_COLUMN_SOURCE_FILE:   return muscleCompare(myInput->GetFileName(),        hisInput->GetFileName());
      case MOVE_COLUMN_SOURCE_FORMAT: return muscleCompare(myInput->GetInputFileFormat(), hisInput->GetInputFileFormat());
      case MOVE_COLUMN_SOURCE_RATE:   return muscleCompare(myInput->GetInputFileSampleRate(),  hisInput->GetInputFileSampleRate());
      case MOVE_COLUMN_SOURCE_WIDTH:  return muscleCompare(myInput->GetInputFileSampleWidth(),  hisInput->GetInputFileSampleWidth());
      case MOVE_COLUMN_LENGTH:        return muscleCompare(myInput->GetInputFileDurationMicros(), hisInput->GetInputFileDurationMicros());
      case MOVE_COLUMN_CHANNELS:      return muscleCompare(myInput->GetFileStreams(),     hisInput->GetFileStreams());
      case MOVE_COLUMN_DEST_FILE:     return muscleCompare(myOutput->GetFileName(),       hisOutput->GetFileName());
      case MOVE_COLUMN_DEST_FORMAT:   return muscleCompare(myOutput->GetOutputFileFormat(),hisOutput->GetOutputFileFormat());
      case MOVE_COLUMN_DEST_RATE:     return muscleCompare(myOutput->GetOutputFileSampleRate(), hisOutput->GetOutputFileSampleRate());
      case MOVE_COLUMN_DEST_WIDTH:    return muscleCompare(myOutput->GetOutputFileSampleWidth(), hisOutput->GetOutputFileSampleWidth());
      case MOVE_COLUMN_QUALITY:       return muscleCompare(myConv->GetConversionQuality(), hisConv->GetConversionQuality());
      case MOVE_COLUMN_PERCENT_DONE:  return muscleCompare(GetPercentDone(),              rhs->GetPercentDone());
      case MOVE_COLUMN_STATUS:        
      {
         int ret = muscleCompare(GetStatus(), rhs->GetStatus());
         return ret ? ret : muscleCompare(_statusString, rhs->_statusString);
      }
   }
   return QListViewItem::compare(i, col, ascending);
}

void AudioMoveItem :: Halt()
{
   if (_status == MOVE_STATUS_PROCESSING) _status = MOVE_STATUS_WAITING;
}

status_t AudioMoveItem :: SendBuffers(bool allowChangeStatus)
{
   if (_status != MOVE_STATUS_ERROR)
   {
      if (_samplesComplete >= _totalSamples) 
      {
         SetStatus(MOVE_STATUS_COMPLETE);  // nothing more to do!
         ShutdownInternalThreads();
      }
      else if (((_status == MOVE_STATUS_PROCESSING)||(_status == MOVE_STATUS_WAITING))&&(_owner->_pause->isChecked() == false))
      {
         static const uint32 MAX_BUFFERS_PER_PROCESS = 5;
         static const uint32 MAX_BUFFER_SIZE_SAMPLES = 64*1024;  // 256KB buffers (i.e. 64K*sizeof(float)) ought to be large enough

         if (allowChangeStatus) SetStatus(MOVE_STATUS_PROCESSING);
         while((_samplesLeft > 0)&&(_numActiveBuffers < MAX_BUFFERS_PER_PROCESS))
         {
            MessageRef msg = GetMessageFromPool(_tag);
            uint32 bufSamples = (uint32) muscleMin((uint64)MAX_BUFFER_SIZE_SAMPLES, _samplesLeft);
            ByteBufferRef buf = GetByteBufferFromPool(bufSamples * sizeof(float));
            if ((msg())&&(buf())                                                                             &&
                ((bufSamples < _samplesLeft)||(msg()->AddBool(AUDIOMOVE_NAME_ISLAST, true) == B_NO_ERROR))   &&
                (msg()->AddInt32(AUDIOMOVE_NAME_ORIGSIZE, bufSamples) == B_NO_ERROR)                         &&
                (msg()->AddFlat(AUDIOMOVE_NAME_BUF, FlatCountableRef(buf.GetGeneric(), true)) == B_NO_ERROR) &&
                (_threads[STAGE_SOURCE]()->MessageReceivedFromUpstream(msg) == B_NO_ERROR)) 
            {
               _samplesLeft -= bufSamples;
               _numActiveBuffers++;
            }
         }
         return (_numActiveBuffers > 0) ? B_NO_ERROR : B_ERROR;
      }
   }
   return B_ERROR;
}

status_t AudioMoveItem :: MessageReceivedFromUpstream(MessageRef msg)
{
   if (msg())
   {
      if (msg()->HasName(AUDIOMOVE_NAME_OPENFILE))
      {
         if (msg()->HasName(AUDIOMOVE_NAME_STATUS))
         {
            SetStatus(MOVE_STATUS_ERROR);
            ShutdownInternalThreads();
         }
      }
      else
      {
         int32 osize;
         if ((msg()->HasName(AUDIOMOVE_NAME_BUF))&&(msg()->FindInt32(AUDIOMOVE_NAME_ORIGSIZE, &osize) == B_NO_ERROR)) _samplesComplete += osize;
         else 
         {
            SetStatus(MOVE_STATUS_ERROR);
            ShutdownInternalThreads();
         }
      }

      const char * s;
      if (msg()->FindString(AUDIOMOVE_NAME_STATUS, &s) == B_NO_ERROR) SetStatusString(ToQ(s));
   }

   _numActiveBuffers--;

   // limit the updates to 5 per second, to avoid spending too much CPU time updating the GUI
   uint64 now = GetRunTime64();
   if (now > _lastUpdateTime+500000)
   {
      _lastUpdateTime = now;
      Update(false);
   }

   if (_status == MOVE_STATUS_PROCESSING) (void) SendBuffers(false);
   return B_NO_ERROR;
}

AudioMoveListView :: AudioMoveListView(QWidget * parent) : QListView(parent) 
{
   setAcceptDrops(true);
   setSelectionMode(Extended);
   setAllColumnsShowFocus(true);
   for (uint32 i=0; i<NUM_MOVE_COLUMNS; i++) 
   {
      addColumn(GetMoveColumnName(i));
      setColumnAlignment(i, ((i == MOVE_COLUMN_SOURCE_FILE)||(i==MOVE_COLUMN_DEST_FILE)) ? AlignRight : AlignHCenter);
      setColumnWidthMode(i, QListView::Manual);
   }
}

QString AudioMoveListView :: GetMoveColumnName(uint32 i) const
{
   switch(i)
   {
      case MOVE_COLUMN_JOB_ID:        return tr("Job #");
      case MOVE_COLUMN_SOURCE_FILE:   return tr("Src File");
      case MOVE_COLUMN_SOURCE_FORMAT: return tr("Src Format");
      case MOVE_COLUMN_SOURCE_RATE:   return tr("Src Rate");
      case MOVE_COLUMN_SOURCE_WIDTH:  return tr("Src Width");
      case MOVE_COLUMN_LENGTH:        return tr("Duration");
      case MOVE_COLUMN_CHANNELS:      return tr("Chans");
      case MOVE_COLUMN_DEST_FILE:     return tr("Dest File");
      case MOVE_COLUMN_DEST_FORMAT:   return tr("Dest Format");
      case MOVE_COLUMN_DEST_RATE:     return tr("Dest Rate");
      case MOVE_COLUMN_DEST_WIDTH:    return tr("Dest Width");
      case MOVE_COLUMN_QUALITY:       return tr("Quality");
      case MOVE_COLUMN_PERCENT_DONE:  return tr("Completed");
      case MOVE_COLUMN_STATUS:        return tr("Status");
      default:                        return tr("Unknown");
   }
}

void AudioMoveListView :: dragEnterEvent(QDragEnterEvent * e)
{
   if (e->provides("text/uri-list")) e->accept();
                                else QListView::dragEnterEvent(e);
}

void AudioMoveListView :: dropEvent(QDropEvent * e)
{
   if (e->provides("text/uri-list"))
      {
      e->accept();
      QStringList list;
      if (QUriDrag::decodeLocalFiles(e,list)) emit FilesDropped(list);
   }
   else QListView::dropEvent(e);
}

AudioMoveWindow :: AudioMoveWindow(const Message & args, QWidget * parent, const char * name, WFlags f) : QMainWindow(parent, name, f), _chooseDestDir(NULL), _addFiles(NULL), _quitOnIdle(false), _forceQuit(false), _numInitializing(0), _tagCounter(0), _audioSetupThread(NULL)
{
   for (uint32 i=0; i<ARRAYITEMS(_audioStatusColorGroupValid); i++) _audioStatusColorGroupValid[i] = false;

   // Setup GUI
   QWidget * cw = new QWidget(this);
   setCentralWidget(cw);

   QVBoxLayout * mainLayout = new QVBoxLayout(cw, 5);
   {
      QGroupBox * settingsGroup = new QGroupBox(cw);
      {
         QVBoxLayout * settingsGroupLayout = new QVBoxLayout(settingsGroup, 3);

         QWidget * lineOne = new QWidget(settingsGroup);
         {
            QHBoxLayout * lineOneLayout = new QHBoxLayout(lineOne, 5, 0);

            _targetFormat = CreateSettingsComboBox(tr("Output Format:"), lineOne, lineOneLayout);
            connect(_targetFormat, SIGNAL(activated(int)), this, SLOT(UpdateButtons()));
            for (uint32 i=0; i<NUM_AUDIO_FORMATS; i++) if (i != AUDIO_FORMAT_NORMALIZED) _targetFormat->insertItem(GetFileFormatName(i));
            AdjustComboBoxWidth(_targetFormat);

            _targetSampleRate = CreateSettingsComboBox(tr("Rate:"), lineOne, lineOneLayout);
            for (uint32 j=0; j<NUM_AUDIO_RATES; j++) _targetSampleRate->insertItem(GetSampleRateName(j, false));
            AdjustComboBoxWidth(_targetSampleRate);

            _targetSampleWidth = CreateSettingsComboBox(tr("Width:"), lineOne, lineOneLayout);
            for (uint32 k=0; k<NUM_AUDIO_WIDTHS; k++) _targetSampleWidth->insertItem(GetSampleWidthName(k, false));
            AdjustComboBoxWidth(_targetSampleWidth);

            _conversionQuality = CreateSettingsComboBox(tr("Quality:"), lineOne, lineOneLayout);
            for (uint32 m=0; m<NUM_AUDIO_CONVERSION_QUALITIES; m++) _conversionQuality->insertItem(GetConversionQualityName(m));
            AdjustComboBoxWidth(_conversionQuality);

            _maxSimultaneous = CreateSettingsComboBox(tr("Threads:"), lineOne, lineOneLayout);
            for (uint32 n=0; n<MAX_SIMULTANEOUS_PROCESSES; n++) _maxSimultaneous->insertItem(ToQ("%1").arg(n+1));
            connect(_maxSimultaneous, SIGNAL(activated(int)), this, SLOT(MaxSimultaneousChanged()));
            AdjustComboBoxWidth(_maxSimultaneous);
         }
         settingsGroupLayout->addWidget(lineOne);

         QWidget * lineTwo = new QWidget(settingsGroup);
         {
            QHBoxLayout * lineTwoLayout = new QHBoxLayout(lineTwo, 5, 0);
            _showDestDialog = AddButton(tr("Destination Folder:"), SLOT(ShowDestDialog()), lineTwo, lineTwoLayout);

            _outputFolderPath = new QLineEdit(lineTwo);
            lineTwoLayout->addWidget(_outputFolderPath);
         }
         settingsGroupLayout->addWidget(lineTwo);
      }
      mainLayout->addWidget(settingsGroup);

      _processList = new AudioMoveListView(cw);
      connect(_processList, SIGNAL(FilesDropped(const QStringList &)), this, SLOT(AddFiles(const QStringList &)));
      connect(_processList, SIGNAL(selectionChanged()), this, SLOT(UpdateButtons()));
      mainLayout->addWidget(_processList, 10);

      QGroupBox * buttonsGroup = new QGroupBox(cw);
      {
         QHBoxLayout * buttonsLayout = new QHBoxLayout(buttonsGroup, 5);

         buttonsLayout->addStretch();

         _pause = new QCheckBox(tr("Pause"), buttonsGroup);
         connect(_pause, SIGNAL(stateChanged(int)), this, SLOT(TogglePaused()));
         buttonsLayout->addWidget(_pause);

         buttonsLayout->addStretch();
         _showAddFilesDialog = AddButton(tr("Add Files..."), SLOT(ShowAddFilesDialog()), buttonsGroup, buttonsLayout);

         buttonsLayout->addStretch();
         _removeSelected = AddButton(tr("Remove Selected Transfers"), SLOT(RemoveSelected()), buttonsGroup, buttonsLayout);
         buttonsLayout->addStretch();
         _removeComplete = AddButton(tr("Remove Completed/Aborted Transfers"), SLOT(RemoveComplete()), buttonsGroup, buttonsLayout);
         buttonsLayout->addStretch();
      }
      mainLayout->addWidget(buttonsGroup);
   }

   // load initial settings
   {
      String dest  = FromQ(QDir::home().absPath());
      String addFilesDir;
      int32 format = AUDIO_FORMAT_SOURCE;
      int32 rate   = AUDIO_RATE_48000;
      int32 width  = AUDIO_WIDTH_FLOAT;
      int32 maxp   = 3;
      int32 cq     = DEFAULT_AUDIO_CONVERSION_QUALITY;

      Message settingsMsg;
      if (LoadMessageFromRegistry("audiomove", settingsMsg) == B_NO_ERROR)
      {
         (void) settingsMsg.FindString("amw_dest",    dest);
         (void) settingsMsg.FindInt32("amw_tformat", &format);
         (void) settingsMsg.FindInt32("amw_trate",   &rate);
         (void) settingsMsg.FindInt32("amw_twidth",  &width);
         (void) settingsMsg.FindInt32("amw_qual",    &cq);
         (void) settingsMsg.FindInt32("amw_maxp",    &maxp);
         if (settingsMsg.FindString("amw_afd", addFilesDir) == B_NO_ERROR) _addFilesDir = ToQ(addFilesDir());
         (void) RestoreWindowPositionFromArchive(this, settingsMsg);
      }

      int32 colw;
      for (uint32 i=0; settingsMsg.FindInt32("amw_colw", i, &colw) == B_NO_ERROR; i++) _processList->setColumnWidth(i, colw);

      SetDestination(dest);
      SetTargetFormat(format);
      SetTargetSampleRate(rate);
      SetTargetSampleWidth(width);
      SetConversionQuality(cq);
      SetMaxProcesses(maxp);
   }

   // Allow us to override our settings from the command line 
   {
      String temp;

      if ((args.FindString("quit", temp) == B_NO_ERROR)||(args.FindString("autoquit", temp) == B_NO_ERROR))
      {
         _quitOnIdle = ParseBool(temp, true);
         if (temp.ToLowerCase() == "force") _forceQuit = _quitOnIdle = true;
      }
      if (((args.FindString("pause", temp) == B_NO_ERROR)||(args.FindString("paused", temp) == B_NO_ERROR))&&(temp.Length() > 0)) _pause->setChecked(ParseBool(temp, true));
      if ((args.FindString("dir", temp) == B_NO_ERROR)||(args.FindString("dest", temp) == B_NO_ERROR)||(args.FindString("destdir", temp) == B_NO_ERROR)) SetDestination(temp);
      if ((args.FindString("format", temp) == B_NO_ERROR)||(args.FindString("destformat", temp) == B_NO_ERROR)) SetTargetFormat(GetFileFormatByName(ToQ(temp()), AUDIO_FORMAT_SOURCE));
      if ((args.FindString("rate", temp) == B_NO_ERROR)||(args.FindString("destrate", temp) == B_NO_ERROR)) SetTargetSampleRate(GetSampleRateByName(ToQ(temp()), AUDIO_RATE_SOURCE));
      if ((args.FindString("width", temp) == B_NO_ERROR)||(args.FindString("destwidth", temp) == B_NO_ERROR)) SetTargetSampleWidth(GetSampleWidthByName(ToQ(temp()), AUDIO_WIDTH_SOURCE));
      if ((args.FindString("threads", temp) == B_NO_ERROR)||(args.FindString("maxthreads", temp) == B_NO_ERROR)) SetMaxProcesses(muscleClamp(atoi(temp()), 1, 9));
      if ((args.FindString("qual", temp) == B_NO_ERROR)||(args.FindString("quality", temp) == B_NO_ERROR)||(args.FindString("conversionquality", temp) == B_NO_ERROR)) SetConversionQuality(GetConversionQualityByName(ToQ(temp()), DEFAULT_AUDIO_CONVERSION_QUALITY));

      for (uint32 i=0; args.FindString("file", i, temp) == B_NO_ERROR; i++) AddFile(ToQ(temp()));
   }

   QTimer::singleShot(1, this, SLOT(UpdateButtons()));  // do this after event loop starts, in case we need to close() ASAP
}

AudioMoveWindow :: ~AudioMoveWindow()
{
   if (_audioSetupThread)
   {
      _audioSetupThread->ShutdownInternalThread();
      delete _audioSetupThread;
   }

   Message settingsMsg; (void) SaveWindowPositionToArchive(this, settingsMsg);

   for (uint32 i=0; i<NUM_MOVE_COLUMNS; i++) settingsMsg.AddInt32("amw_colw", _processList->columnWidth(i));

   settingsMsg.AddString("amw_dest",   GetDestination());
   settingsMsg.AddInt32("amw_tformat", GetTargetFormat());
   settingsMsg.AddInt32("amw_trate",   GetTargetSampleRate());
   settingsMsg.AddInt32("amw_twidth",  GetTargetSampleWidth());
   settingsMsg.AddInt32("amw_qual",    GetConversionQuality());
   settingsMsg.AddInt32("amw_maxp",    GetMaxProcesses());
   settingsMsg.AddString("amw_afd",    FromQ(_addFilesDir));

   (void) SaveMessageToRegistry("audiomove", settingsMsg);
}

QComboBox * AudioMoveWindow :: CreateSettingsComboBox(const QString & label, QWidget * parent, QBoxLayout * layout)
{
   QComboBox * ret;
   QWidget * subWidget = new QWidget(parent);
   {
      QHBoxLayout * subLayout = new QHBoxLayout(subWidget, 3);

      subLayout->addStretch(1);
      subLayout->addWidget(new QLabel(label, subWidget));

      ret = new QComboBox(subWidget);
      subLayout->addWidget(ret);
      subLayout->addStretch(1);
   }
   layout->addWidget(subWidget); 
   return ret;
}

QPushButton * AudioMoveWindow :: AddButton(const QString & label, const char * slot, QWidget * parent, QBoxLayout * layout)
{
   QPushButton * ret = new QPushButton(label, parent);
   connect(ret, SIGNAL(clicked()), this, slot);
   layout->addWidget(ret);
   return ret;
}

uint32 AudioMoveWindow :: GetNumActiveTransfers(bool selectedOnly, uint32 * optRetNumSel, uint32 * optRetErrs) const
{
   if (optRetNumSel) *optRetNumSel = 0;
   if (optRetErrs)   *optRetErrs   = 0;

   // See we have any non-completed, non-error'd events
   uint32 numActive = 0;
   HashtableIterator<uint32, AudioMoveItem *> iter(_moveItems);
   AudioMoveItem * next;
   while(iter.GetNextValue(next) == B_NO_ERROR)
   {
      if ((optRetNumSel)&&(next->isSelected())) (*optRetNumSel)++;
      switch(next->GetStatus())
      { 
         case MOVE_STATUS_ERROR:
            if (optRetErrs) (*optRetErrs)++;
         break;

         case MOVE_STATUS_COMPLETE:
            // do nothing
         break;

         default:
            if ((selectedOnly == false)||(next->isSelected())) numActive++;
         break;
      }
   }
   return numActive;
}

void AudioMoveWindow :: closeEvent(QCloseEvent * e)
{
   uint32 numActive = GetNumActiveTransfers(false, NULL, NULL);
   if ((numActive > 0)&&(QMessageBox::warning(this, tr("Incomplete Transfers Warning"), tr("There are %1 transfers still in progress.  Are you sure you want to quit now?").arg(numActive), QMessageBox::Yes, QMessageBox::No) == QMessageBox::No))
   {
      e->ignore();
      return;
   }
   QMainWindow::closeEvent(e);
}

void AudioMoveWindow :: SetDestination(const String & d)
{
   _outputFolderPath->setText(ToQ(d()));
}

String AudioMoveWindow :: GetDestination() const
{
   return FromQ(_outputFolderPath->text());
}

void AudioMoveWindow :: SetTargetFormat(uint32 format)
{
   _targetFormat->setCurrentItem(format);
}

uint32 AudioMoveWindow :: GetTargetFormat() const
{
   return _targetFormat->currentItem();
}

QString AudioMoveWindow :: GetFileDurationName(int64 frames, uint32 samplesPerSecond) const
{
   if (samplesPerSecond > 0)
   {
      //uint64 numSamples = frames%samplesPerSecond;
      uint64 numSeconds = frames/samplesPerSecond;
      uint64 numMinutes = numSeconds/60;  numSeconds = numSeconds%60;
      uint64 numHours   = numMinutes/60;  numMinutes = numMinutes%60;
      char buf[256]; sprintf(buf, "%lluh %02llum %02llus", numHours, numMinutes, numSeconds);
      const char * s = buf;
      while((*s != '\0')&&(muscleInRange(*s, '1', '9') == false)) s++;
      return (*s ? ToQ(s) : ToQ("--"));
   }
   return ToQ("");
}

QString AudioMoveWindow :: GetFileFormatName(uint32 format) const
{
   switch(format)
   {
      case AUDIO_FORMAT_SOURCE:     return tr("Source");
      case AUDIO_FORMAT_WAV:        return tr(".WAV");
      case AUDIO_FORMAT_AIFF:       return tr(".AIFF");
      case AUDIO_FORMAT_LCSRAW:     return tr(".LCS");
      case AUDIO_FORMAT_NORMALIZED: return ToQ("");
      default:                      return tr("Unknown");
   }
}

uint32 AudioMoveWindow :: GetFileFormatByName(const QString & name, uint32 defaultFormat) const
{
   for (uint32 i=0; i<NUM_AUDIO_FORMATS; i++) if (GetFileFormatName(i).contains(name, false)) return i;
   return defaultFormat;
}

void AudioMoveWindow :: SetTargetSampleRate(uint32 rate)
{
   _targetSampleRate->setCurrentItem(rate);
}

uint32 AudioMoveWindow :: GetTargetSampleRate() const
{
   return _targetSampleRate->currentItem();
}

QString AudioMoveWindow :: GetSampleRateName(uint32 rate, bool inTable) const
{
   switch(rate)
   {
      case AUDIO_RATE_SOURCE: return inTable ? ToQ("") : tr("Source");
      case AUDIO_RATE_48000:  return tr("48KHz");
      case AUDIO_RATE_44100:  return tr("44.1KHz");
      case AUDIO_RATE_22050:  return tr("22.05KHz");
      default:                return tr("%1KHz").arg(rate/1000, 0, 'g', 2);
   }
}

uint32 AudioMoveWindow :: GetSampleRateByName(const QString & name, uint32 defaultRate) const
{
   for (uint32 i=0; i<NUM_AUDIO_RATES; i++) if (GetSampleRateName(i, false).contains(name, false)) return i;
   return defaultRate;
}


void AudioMoveWindow :: SetTargetSampleWidth(uint32 width)
{
   _targetSampleWidth->setCurrentItem(width);
}

uint32 AudioMoveWindow :: GetTargetSampleWidth() const
{
   return _targetSampleWidth->currentItem();
}

QString AudioMoveWindow :: GetSampleWidthName(uint32 width, bool inTable) const
{
   switch(width)
   {
      case AUDIO_WIDTH_SOURCE: return inTable ? ToQ("") : tr("Source");
      case AUDIO_WIDTH_FLOAT:  return tr("Float");
      case AUDIO_WIDTH_INT32:  return tr("32 bit");
      case AUDIO_WIDTH_INT24:  return tr("24 bit");
      case AUDIO_WIDTH_INT16:  return tr("16 bit");
      case AUDIO_WIDTH_INT8:   return tr("8 bit");
      default:                 return tr("Unknown");
   }
}

uint32 AudioMoveWindow :: GetSampleWidthByName(const QString & name, uint32 defaultWidth) const
{
   for (uint32 i=0; i<NUM_AUDIO_WIDTHS; i++) if (GetSampleWidthName(i, false).contains(name, false)) return i;
   return defaultWidth;
}

void AudioMoveWindow :: SetConversionQuality(uint32 qual)
{
   _conversionQuality->setCurrentItem(qual);
}

uint32 AudioMoveWindow :: GetConversionQuality() const
{
   return _conversionQuality->currentItem();
}

QString AudioMoveWindow :: GetConversionQualityName(uint32 qual) const
{
   switch(qual)
   {
      case AUDIO_CONVERSION_QUALITY_BEST:   return tr("Best");
      case AUDIO_CONVERSION_QUALITY_BETTER: return tr("Better");
      case AUDIO_CONVERSION_QUALITY_GOOD:   return tr("Good");
      default:                              return tr("Unknown");
   }
}

uint32 AudioMoveWindow :: GetConversionQualityByName(const QString & name, uint32 defaultQual) const
{
   for (uint32 i=0; i<NUM_AUDIO_CONVERSION_QUALITIES; i++) if (GetConversionQualityName(i).contains(name, false)) return i;
   return defaultQual;
}

QString AudioMoveWindow :: GetStatusName(uint32 status) const
{
   switch(status)
   {
      case MOVE_STATUS_WAITING:    return tr("Waiting");
      case MOVE_STATUS_PROCESSING: return _pause->isChecked() ? tr("Paused") : tr("Processing");
      case MOVE_STATUS_COMPLETE:   return tr("Finished");
      case MOVE_STATUS_ERROR:      return tr("Error");
      default:                     return tr("Unknown");
   }
}

void AudioMoveWindow :: MaxSimultaneousChanged()
{
   // First, go through and set all active processes to the "waiting" state...
   HashtableIterator<uint32, AudioMoveItem *> iter(_moveItems);
   AudioMoveItem * next;
   while(iter.GetNextValue(next) == B_NO_ERROR) next->Halt();

   // Then Dequeue as many processes as can now fit
   DequeueTransfers();
   ForceUpdateAll();
}

void AudioMoveWindow :: SetMaxProcesses(uint32 maxp)
{
   _maxSimultaneous->setCurrentItem(maxp-1);
}

uint32 AudioMoveWindow :: GetMaxProcesses() const
{
   return _maxSimultaneous->currentItem()+1;
}

void AudioMoveWindow :: ShowAddFilesDialog()
{
   if (_addFiles == NULL) 
   {
      _addFiles = new QFileDialog(QString::null, QString::null, this, NULL, true);  // made modal to work around Qt/OS/X bug --jaf
      _addFiles->setMode(QFileDialog::ExistingFiles);
      connect(_addFiles, SIGNAL(fileSelected(const QString &)), this, SLOT(DialogAddFile(const QString &)));
      connect(_addFiles, SIGNAL(filesSelected(const QStringList &)), this, SLOT(DialogAddFiles(const QStringList &)));
   }
   _addFiles->setSelection(ToQ(""));
   _addFiles->setDir(_addFilesDir);
   _addFiles->rereadDir();
   _addFiles->show();
}

void AudioMoveWindow :: DialogAddFiles(const QStringList & files)
{
   _addFilesDir = _addFiles->dirPath();
   AddFiles(files);
}

void AudioMoveWindow :: DialogAddFile(const QString & fn)
{
   _addFilesDir = _addFiles->dirPath();
   AddFile(fn);
}

void AudioMoveWindow :: AddFiles(const QStringList & files)
{
   int numFiles = files.size();
   for (int i=0; i<numFiles; i++) AddFile(files[i]);
}

void AudioMoveWindow :: AddFile(const QString & qfn)
{
   String fn = FromQ(qfn);
   MessageRef setupMsg = GetMessageFromPool(SETUP_COMMAND_SETUP);
   if ((setupMsg())&&
       (setupMsg()->AddString(SETUP_NAME_SOURCEFILE, FromQ(qfn))             == B_NO_ERROR)&&
       (setupMsg()->AddString(SETUP_NAME_DESTDIR, GetDestination())          == B_NO_ERROR)&&
       (setupMsg()->AddInt32(SETUP_NAME_TARGETFORMAT, GetTargetFormat())     == B_NO_ERROR)&&
       (setupMsg()->AddInt32(SETUP_NAME_TARGETRATE, GetTargetSampleRate())   == B_NO_ERROR)&&
       (setupMsg()->AddInt32(SETUP_NAME_TARGETWIDTH, GetTargetSampleWidth()) == B_NO_ERROR)&&
       (setupMsg()->AddInt32(SETUP_NAME_QUALITY, GetConversionQuality())     == B_NO_ERROR))
   {
      // demand-allocate the setup thread
      if (_audioSetupThread == NULL)
      {
         _audioSetupThread = new AudioSetupThread(this);
         if (_audioSetupThread->StartInternalThread() != B_NO_ERROR)
         {
            LogTime(MUSCLE_LOG_ERROR, "Couldn't start audio setup thread!\n");
            delete _audioSetupThread;
            _audioSetupThread = NULL;
         }
      }
      if ((_audioSetupThread)&&(_audioSetupThread->SendMessageToInternalThread(setupMsg) == B_NO_ERROR)) _numInitializing++;
   }
}


void AudioMoveWindow :: DequeueTransfers()
{
   if (_pause->isChecked() == false)
   {
      uint32 numActive = 0;
      HashtableIterator<uint32, AudioMoveItem *> iter(_moveItems);
      AudioMoveItem * next;
      uint32 maxNumActive = GetMaxProcesses();
      while((numActive < maxNumActive)&&(iter.GetNextValue(next) == B_NO_ERROR))
      {
         bool update = false;
         switch(next->GetStatus())
         {
            case MOVE_STATUS_WAITING: 
               update = true;
               (void) next->StartInternalThreads();  // only has an effect the first time
            // fall thru!
            case MOVE_STATUS_PROCESSING:
               if (next->SendBuffers(true) == B_NO_ERROR) numActive++;
               if (update) next->Update(true);
            break;
         }
      }
   }
}

void AudioMoveWindow :: ShowDestDialog()
{
   if (_chooseDestDir == NULL) 
   {
      _chooseDestDir = new ChooseFolderFileDialog(this);
      connect(_chooseDestDir, SIGNAL(FolderSelected(const QString &)), _outputFolderPath, SLOT(setText(const QString &)));
   }
   _chooseDestDir->setSelection(ToQ(""));
   _chooseDestDir->setDir(_outputFolderPath->text());
   _chooseDestDir->rereadDir();
   _chooseDestDir->show();
}

void AudioMoveWindow :: TogglePaused()
{
   DequeueTransfers();
   ForceUpdateAll();
}

void AudioMoveWindow :: ForceUpdateAll()
{
   // Force update of all active items, so that "Paused" will become "Processing" or vice versa
   HashtableIterator<uint32, AudioMoveItem *> iter(_moveItems);
   AudioMoveItem * next;
   while(iter.GetNextValue(next) == B_NO_ERROR) next->Update(true);
}

void AudioMoveWindow :: RemoveComplete()
{
   HashtableIterator<uint32, AudioMoveItem *> iter(_moveItems);
   uint32 nextKey;
   AudioMoveItem * next;
   while(iter.GetNextKeyAndValue(nextKey, next) == B_NO_ERROR)
   {
      switch(next->GetStatus())
      {
         case MOVE_STATUS_ERROR: case MOVE_STATUS_COMPLETE:
            _moveItems.Remove(nextKey);
            delete next;
         break;
      }
   }
   UpdateButtons();
}

void AudioMoveWindow :: RemoveSelected()
{
   uint32 numSel;
   uint32 numActive = GetNumActiveTransfers(true, &numSel, NULL);
   if ((numActive == 0)||(QMessageBox::warning(this, tr("Incomplete Transfers Warning"), tr("%1 of the %2 selected transfers are still in progress.  Are you sure you want to remove them now?").arg(numActive).arg(numSel), QMessageBox::Yes, QMessageBox::No) == QMessageBox::Yes))
   {
      HashtableIterator<uint32, AudioMoveItem *> iter(_moveItems);
      uint32 nextKey;
      AudioMoveItem * next;
      while(iter.GetNextKeyAndValue(nextKey, next) == B_NO_ERROR)
      {
         if (next->isSelected())
         {
            _moveItems.Remove(nextKey);
            delete next;
         }
      }
      UpdateButtons();
   }
}

void AudioMoveWindow :: UpdateButtons()
{
   uint32 numSel, numErrs;
   uint32 numActive = GetNumActiveTransfers(false, &numSel, &numErrs);
   _removeSelected->setEnabled(numSel > 0);
   _removeComplete->setEnabled(numActive < _moveItems.GetNumItems());
   _targetSampleRate->setEnabled( GetTargetFormat() != AUDIO_FORMAT_LCSRAW);  // LCS native only supports 48KHz
   _targetSampleWidth->setEnabled(GetTargetFormat() != AUDIO_FORMAT_LCSRAW);  // LCS native only supports float

   if ((_numInitializing == 0)&&(_quitOnIdle)&&(numActive == 0))
   {
      // do not combine these if statements!
      if (_forceQuit) close(false);
      else
      {
         if (numErrs == 0) close(false);
                      else _quitOnIdle = false;
      }
   }
}

status_t AudioMoveWindow :: MessageReceivedFromUpstream(MessageRef msg)
{
   // We can't handle it here, since we are in an output thread... so we'll
   // post it asynchronously back to the main GUI thread, instead!
   QApplication::postEvent(this, new BufferReturnedEvent(msg, false));
   return B_NO_ERROR;
}

bool AudioMoveWindow :: event(QEvent * evt)
{
   switch(evt->type())
   {
      case BUFFER_RETURNED_EVENT_TYPE:
      {
         const BufferReturnedEvent * bre = dynamic_cast<BufferReturnedEvent *>(evt);
         if (bre) 
         {
            MessageRef msg = bre->GetMessage();
            if (msg())
            {
               if (bre->IsFromSetup())
               {
                  _numInitializing--;

                  MessageRef subMsg;
                  for (int32 i=0; msg()->FindMessage(SETUP_NAME_SETUPRESULT, i, subMsg) == B_NO_ERROR; i++)
                  {
                     String errorString; (void) subMsg()->FindString(SETUP_NAME_ERROR, errorString);

                     GenericRef itag, otag;
                     AudioMoveThreadRef inputThread, outputThread;
                     uint32 quality;
                     if ((subMsg()->FindTag(SETUP_NAME_INPUTTHREAD, itag) == B_NO_ERROR)&&(subMsg()->FindTag(SETUP_NAME_OUTPUTTHREAD, otag) == B_NO_ERROR)&&(subMsg()->FindInt32(SETUP_NAME_QUALITY, (int32*)&quality) == B_NO_ERROR)&&(inputThread.SetFromGeneric(itag) == B_NO_ERROR)&&(outputThread.SetFromGeneric(otag) == B_NO_ERROR))
                     {
                        QListViewItem * last = _processList->firstChild();
                        while((last)&&(last->nextSibling())) last = last->nextSibling();

                        AudioMoveItem * ami = new AudioMoveItem(++_tagCounter, this, _processList, last, inputThread, AudioMoveThreadRef(new SampleRateThread(GetSampleRateValueFromCode(inputThread()->GetOutputFileSampleRate()), GetSampleRateValueFromCode(outputThread()->GetInputFileSampleRate()), quality, inputThread()->GetFileStreams()), NULL), outputThread);
                        if (_moveItems.Put(ami->GetTag(), ami) == B_NO_ERROR)
                        {
                           if (errorString.Length() > 0) 
                           {
                              ami->SetStatus(MOVE_STATUS_ERROR);
                              ami->SetStatusString(ToQ(errorString()));
                           }
                           ami->Update(true);
                        }
                     }
                  }
                  DequeueTransfers();
               }
               else
               {
                  AudioMoveItem * item;
                  if (_moveItems.Get(msg()->what, item) == B_NO_ERROR)
                  {
                     item->MessageReceivedFromUpstream(msg);
                     if (item->GetStatus() != MOVE_STATUS_PROCESSING) DequeueTransfers();
                  }
               }
            }
         }
      }
      return true;

      default:
         return QMainWindow::event(evt);
   }
}

const QColorGroup & AudioMoveWindow :: GetStatusColorGroup(uint32 status, const QColorGroup & cg) const
{
   bool & v = _audioStatusColorGroupValid[status];
   QColorGroup & scg = _audioStatusColorGroups[status];
   if (v == false)
   {
      QColor color;
      switch(status)
      {
         case MOVE_STATUS_WAITING:    color = QColor(200,200,255); break;
         case MOVE_STATUS_PROCESSING: color = QColor(255,255,200); break;
         case MOVE_STATUS_COMPLETE:   color = QColor(200,255,200); break;
         case MOVE_STATUS_ERROR:      color = QColor(255,200,200); break;
      }
      scg = cg;
      scg.setColor(QColorGroup::Base, color);
      scg.setColor(QColorGroup::Highlight, color.dark(120));
      scg.setColor(QColorGroup::Text, black);
      scg.setColor(QColorGroup::HighlightedText, black);
      v = true;
   }
   return scg;
}

uint32 GetSampleRateValueFromCode(uint32 rateCode)
{
   switch(rateCode)
   {
      case AUDIO_RATE_48000: return 48000;
      case AUDIO_RATE_44100: return 44100;
      case AUDIO_RATE_22050: return 22050;
      default:               return rateCode;
   }
}

uint32 GetSampleRateCodeFromValue(uint32 value)
{
   switch(value)
   {
      case 48000: return AUDIO_RATE_48000;
      case 44100: return AUDIO_RATE_44100;
      case 22050: return AUDIO_RATE_22050;
      default:    return NUM_AUDIO_RATES;
   }
}

};  // end namespace audiomove
