/*
** Copyright (C) 2004 Level Control Systems <jaf@lcsaudio.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef AudioMoveWindow_h
#define AudioMoveWindow_h

#include <qfiledialog.h>
#include <qlistview.h>
#include <qmainwindow.h>
#include <qdragobject.h>
#include <qpalette.h>
#include "audiomove/AudioMoveNameSpace.h"
#include "audiomove/AudioMoveThread.h"
#include "util/String.h"

class QCheckBox;
class QComboBox;
class QLineEdit;
class QPushButton;

namespace audiomove {

class AudioMoveThread;
class AudioMoveWindow;
class AudioSetupThread;

enum {
   MOVE_STATUS_WAITING = 0,
   MOVE_STATUS_PROCESSING,
   MOVE_STATUS_COMPLETE,
   MOVE_STATUS_ERROR,
   NUM_MOVE_STATUSES
};

/** Apparently QFileDialog is incapable of giving us what we want on its own :^P */
class ChooseFolderFileDialog : public QFileDialog
{
Q_OBJECT

public:
   ChooseFolderFileDialog(QWidget * parent) : QFileDialog(parent) 
   {
      setMode(QFileDialog::Directory);
   }

   virtual void accept()
   {
      QFileDialog::accept();
      emit FolderSelected(((selectedFile().length() > 0)&&(QDir(selectedFile()).exists()) ? selectedFile() : dirPath()));
   }

signals:
   void FolderSelected(const QString & path);
};

class AudioMoveListView : public QListView
{
Q_OBJECT

public:
   AudioMoveListView(QWidget * parent);
 
   virtual void dragEnterEvent(QDragEnterEvent * e);
   virtual void dropEvent(QDropEvent * e);

signals:
   void FilesDropped(const QStringList & list);

private:
   QString GetMoveColumnName(uint32 col) const;
};

/** Objects of this class represent a transfer -- both the UI object and the implementation. */
class AudioMoveItem : public QListViewItem, public IMessageAcceptor
{
public:
   AudioMoveItem(uint32 tag, AudioMoveWindow * owner, QListView * lv, QListViewItem * after, AudioMoveThreadRef input, AudioMoveThreadRef convert, AudioMoveThreadRef output);
   virtual ~AudioMoveItem();

   status_t StartInternalThreads();
   void ShutdownInternalThreads();

   /** Updates all our cells appropriately. */
   void Update(bool updateAllColumns);

   float GetPercentDone() const {return (_totalSamples > 0) ? muscleClamp((float)(((double)(int64)_samplesComplete)/((double)(int64)_totalSamples)), 0.0f, 1.0f) : 0.0f;}
   uint32 GetStatus() const {return _status;}

   virtual int compare(QListViewItem * i, int col, bool ascending) const;

   uint32 GetTag() const {return _tag;}

   /* If we are currently in processing mode, this will put us back to waiting mode */
   void Halt();

   /** Sends out buffers until are maximum number of buffers are in play */
   status_t SendBuffers(bool allowChangeStatus);

   /** Called by the window object, in the main thread, when a buffer has completed the circuit. */
   virtual status_t MessageReceivedFromUpstream(MessageRef msg);

   /** Overridden to color our cells pretty colors */
   virtual void paintCell(QPainter * p, const QColorGroup & cg, int column, int width, int align);

   void SetStatusString(const QString & ss);
   void SetStatus(uint32 s);

private:
   enum {
      STAGE_SOURCE = 0,
      STAGE_CONVERSION,
      STAGE_DESTINATION,
      NUM_STAGES
   };

   const uint32 _tag;
   bool _internalThreadsStarted;
   QString _statusString;
   uint32 _status;
   uint64 _samplesLeft;
   uint64 _samplesComplete;
   uint64 _totalSamples;
   uint32 _numActiveBuffers;
   AudioMoveWindow * _owner;
   AudioMoveThreadRef _threads[NUM_STAGES];
   uint64 _lastUpdateTime;
};

/** Window class for the AudioMove GUI.  */
class AudioMoveWindow : public QMainWindow, public IMessageAcceptor
{
Q_OBJECT

public:
   /** Standard constructor; all parameters are passed through to the QMainWindow ctor. **/
   AudioMoveWindow(const Message & args, QWidget * parent = NULL, const char * name = NULL, WFlags f = WType_TopLevel);

   /** Standard destructor */
   virtual ~AudioMoveWindow();

   /** Overridden so we can ask the user if he is sure he wants to quit while transfers are running. */
   virtual void closeEvent(QCloseEvent * e);

   /** Called whenever some data has finished processing.  NOTE:  called from various threads!! */
   virtual status_t MessageReceivedFromUpstream(MessageRef msg);

   /** Overridden to handle incoming BufferReturnedEvents */
   virtual bool event(QEvent * evt);

private slots:
   void ShowAddFilesDialog();
   void ShowDestDialog();
   void TogglePaused();
   void RemoveSelected();
   void RemoveComplete();
   void AddFiles(const QStringList & files);
   void AddFile(const QString & file);
   void DialogAddFile(const QString & fn);
   void DialogAddFiles(const QStringList & files);
   void UpdateButtons();
   void MaxSimultaneousChanged();

private:
   friend class AudioMoveItem;

   QComboBox * CreateSettingsComboBox(const QString & label, QWidget * parent, QBoxLayout * layout);
   void RemoveItems(bool selOnly);

   void SetDestination(const String & d);
   String GetDestination() const;

   void SetTargetFormat(uint32 format);
   uint32 GetTargetFormat() const;

   QString GetFileFormatName(uint32 format) const;
   uint32 GetFileFormatByName(const QString & name, uint32 defaultFormat) const;
   const char * GetAudioFormatExtensions(uint32 format) const;
   String ReplaceExtension(const String & ss, const String & ext);

   void SetTargetSampleRate(uint32 rate);
   uint32 GetTargetSampleRate() const;
   QString GetSampleRateName(uint32 rate, bool inTable) const;
   uint32 GetSampleRateByName(const QString & name, uint32 defaultRate) const;

   void SetTargetSampleWidth(uint32 width);
   uint32 GetTargetSampleWidth() const;
   QString GetSampleWidthName(uint32 width, bool inTable) const;
   uint32 GetSampleWidthByName(const QString & name, uint32 defaultWidth) const;

   void SetConversionQuality(uint32 qual);
   uint32 GetConversionQuality() const;
   QString GetConversionQualityName(uint32 width) const;
   uint32 GetConversionQualityByName(const QString & name, uint32 defaultQuality) const;

   void SetMaxProcesses(uint32 maxp);
   uint32 GetMaxProcesses() const;

   QString GetFileDurationName(int64 frames, uint32 samplesPerSecond) const;
   QString GetStatusName(uint32 status) const;

   QPushButton * AddButton(const QString & label, const char * slot, QWidget * parent, QBoxLayout * layout);

   void DequeueTransfers();
   void ForceUpdateAll();

   const QColorGroup & GetStatusColorGroup(uint32 status, const QColorGroup & cg) const;
   uint32 GetNumActiveTransfers(bool selectedOnly, uint32 * optRetNumSel, uint32 * optRetNumErrs) const;

   AudioMoveListView * _processList;
   QPushButton * _showAddFilesDialog;
   QPushButton * _showDestDialog;
   QLineEdit   * _outputFolderPath;
   QComboBox   * _targetFormat;
   QComboBox   * _targetSampleRate;
   QComboBox   * _targetSampleWidth;
   QComboBox   * _conversionQuality;
   QComboBox   * _maxSimultaneous;

   QCheckBox   * _pause;
   QPushButton * _removeSelected;
   QPushButton * _removeComplete;

   ChooseFolderFileDialog * _chooseDestDir;
   QFileDialog * _addFiles;
   QString _addFilesDir;
   bool _quitOnIdle;
   bool _forceQuit;
   uint32 _numInitializing;

   uint32 _tagCounter;
   Hashtable<uint32, AudioMoveItem *> _moveItems;

   AudioSetupThread * _audioSetupThread;

   mutable QColorGroup _audioStatusColorGroups[NUM_MOVE_STATUSES];
   mutable bool _audioStatusColorGroupValid[NUM_MOVE_STATUSES];
};

};  // end namespace audiomove

#endif
