/*
** Copyright (C) 2004 Level Control Systems <jaf@lcsaudio.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
*/

#include <fcntl.h>
#include "dataio/FileDescriptorDataIO.h"
#include "audiomove/LibSndFileIOThread.h"
#include "audiomove/MiscFunctions.h"

namespace audiomove {

LibSndFileIOThread :: LibSndFileIOThread(const String & readPath) : _fileName(readPath), _numWriteStreams(AUDIO_STREAMS_SOURCE), _file(NULL), _inputFileFormat(AUDIO_FORMAT_SOURCE), _outputFileFormat(AUDIO_FORMAT_NORMALIZED), _fileSampleRate(AUDIO_RATE_SOURCE), _fileSampleWidth(AUDIO_WIDTH_SOURCE), _numFrames(0), _numStreams(0)
{
   // empty
}

LibSndFileIOThread :: LibSndFileIOThread(const String & writePath, uint32 writeFormat, uint32 writeSampleWidth, uint32 writeSampleRate, uint8 writeNumStreams) : _fileName(writePath), _numWriteStreams(writeNumStreams), _file(NULL), _inputFileFormat(AUDIO_FORMAT_NORMALIZED), _outputFileFormat(writeFormat), _fileSampleRate(writeSampleRate), _fileSampleWidth(writeSampleWidth), _numFrames(0), _numStreams(writeNumStreams)
{
   // empty
}

LibSndFileIOThread :: ~LibSndFileIOThread()
{
   CloseFileAux();
}

status_t LibSndFileIOThread :: OpenFile()
{
   CloseFile();  // just in case
   
   SF_INFO info; memset(&info, 0, sizeof(info));
   if (_numWriteStreams != AUDIO_STREAMS_SOURCE)
   {
      // Write mode: If we're going to create the file, then fill in our desired parameters for the file here
      info.samplerate = _fileSampleRate;
      info.channels   = _numWriteStreams;

      switch(_outputFileFormat)
      {
         case AUDIO_FORMAT_WAV:  info.format = SF_FORMAT_WAV  | SF_ENDIAN_BIG; break;
         case AUDIO_FORMAT_AIFF: info.format = SF_FORMAT_AIFF | SF_ENDIAN_BIG; break;
         default:                return B_ERROR;  // we only support WAV and AIFF output... for now anyway
      }

      switch(_fileSampleWidth)
      {
         case AUDIO_WIDTH_FLOAT: info.format |= SF_FORMAT_FLOAT;  break;
         case AUDIO_WIDTH_INT32: info.format |= SF_FORMAT_PCM_32; break;
         case AUDIO_WIDTH_INT24: info.format |= SF_FORMAT_PCM_24; break;
         case AUDIO_WIDTH_INT16: info.format |= SF_FORMAT_PCM_16; break;
         case AUDIO_WIDTH_INT8:  info.format |= SF_FORMAT_PCM_U8; break;
         default:                info.format |= SF_FORMAT_FLOAT;  break;
      }

      // See if we can force libsndfile to open the file as big-endian... since that is ELC-native, we'd prefer that
      EnsureFileFolderExists(_fileName);
      _file = sf_open(_fileName(), SFM_WRITE, &info);
      if (_file == NULL)
      {
         // No luck?  Okay, let's try again, without the big-endian requirement
         info.format &= ~(SF_ENDIAN_BIG);
         _file = sf_open(_fileName(), SFM_WRITE, &info);
      }
      if (_file)
      {
         _numFrames = 0;  // no data written, yet!
         return B_NO_ERROR;
      }
   }
   else 
   {
      _file = sf_open(_fileName(), SFM_READ, &info);
      if (_file)
      {
         // Read mode:  extract attributes from libsndfile
              if (info.format & SF_FORMAT_WAV)  _inputFileFormat = AUDIO_FORMAT_WAV;
         else if (info.format & SF_FORMAT_AIFF) _inputFileFormat = AUDIO_FORMAT_AIFF;
         else                                   _inputFileFormat = NUM_AUDIO_FORMATS;  // must be some other format that libsndfile knows, but we don't!

         _fileSampleRate   = info.samplerate;
         _numFrames        = info.frames;
         _numStreams       = info.channels;

         switch(info.format & SF_FORMAT_SUBMASK)
         {
            case SF_FORMAT_FLOAT: case SF_FORMAT_DOUBLE:  
               _fileSampleWidth = AUDIO_WIDTH_FLOAT; 
            break;

            case SF_FORMAT_PCM_32: _fileSampleWidth = AUDIO_WIDTH_INT32; break;
            case SF_FORMAT_PCM_24: _fileSampleWidth = AUDIO_WIDTH_INT24; break;
            case SF_FORMAT_PCM_16: _fileSampleWidth = AUDIO_WIDTH_INT16; break;

            case SF_FORMAT_PCM_S8: case SF_FORMAT_PCM_U8:
               _fileSampleWidth = AUDIO_WIDTH_INT8;  
            break;

            default:
               _fileSampleWidth = NUM_AUDIO_WIDTHS;
            break;
         }
         return B_NO_ERROR;
      }
   }
   CloseFile();

   return B_ERROR;
}

void LibSndFileIOThread :: CloseFile()
{
   CloseFileAux();
}

void LibSndFileIOThread :: CloseFileAux()
{
   if (_file)
   {
      sf_close(_file);
      _file = NULL;
   }
}

ByteBufferRef LibSndFileIOThread :: ProcessBuffer(ByteBufferRef buf, QString & retErrStr, bool isLastBuffer)
{
   if ((_file)&&(buf()))
   {
      const uint32 numBytes = buf()->GetNumBytes();

      if ((numBytes % sizeof(float)) == 0)
      {
         float * samples = (float *) buf()->GetBuffer();
         const uint32 numSamples = numBytes/sizeof(float);

         if (_numWriteStreams == AUDIO_STREAMS_SOURCE)
         {
            // Read from the input file into the buffer
            int32 readSamples = sf_read_float(_file, samples, numSamples);
            if (sf_error(_file) == SF_ERR_NO_ERROR)
            {
               buf()->SetNumBytes(readSamples*sizeof(float), false);
               if (isLastBuffer) CloseFile();  // close file now so that it isn't locked
               return buf;
            }
            else retErrStr = qApp->translate("LibSndFileIOThread", "Problem reading from file");
         }
         else
         {
            if (sf_write_float(_file, samples, numSamples) == numSamples)
            {
               _numFrames += (numSamples / _numStreams);
               if (isLastBuffer) CloseFile();  // close file now so that it is ready for use immediately
               return buf;
            }
            else retErrStr = qApp->translate("LibSndFileIOThread", "Problem writing to file");
         }
      }
      else retErrStr = qApp->translate("LibSndFileIOThread", "Bad chunk length %1").arg(numBytes);
   }
   if (isLastBuffer) CloseFile();  // close file now so that it isn't locked
   return ByteBufferRef();  // error!
}

};  // end namespace audiomove
