/*
** Copyright (C) 2004 Level Control Systems <jaf@lcsaudio.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
*/

#if defined(WIN32)
# include <windows.h>
# include "win32dirent.h"
#else
# include <dirent.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#include <qmessagebox.h>
#include <qapplication.h>
#include <qfile.h>
#include <qcstring.h>
#include <qprocess.h>
#include <qsettings.h>
#include <qdir.h>
#include <qmessagebox.h>
#include <qobjectlist.h>
#include <qcombobox.h>
#include <qdragobject.h>
#include <qdesktopwidget.h>
#include <qbutton.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qmenudata.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qpopupmenu.h>
#include <qsplitter.h>
#include <qtextedit.h>
#include <qcombobox.h>
#include "message/Message.h"
#include "util/ByteBuffer.h"
#include "util/NetworkUtilityFunctions.h"
#include "system/SharedMemory.h"
#include "util/StringTokenizer.h"
#include "util/MiscUtilityFunctions.h"
#include "audiomove/MiscFunctions.h"

// Miscellaneous functions used by AudioMove... 
namespace audiomove {

status_t ReadSettingsFile(const char * prefsPath, Message & settings)
{
   QFile file(ToQ(prefsPath));
   if (file.open(IO_ReadOnly))
   {
      QByteArray array = file.readAll();
      if (array.data())
      {
         ByteBufferRef bufRef = GetByteBufferFromPool(array.size(), (const uint8 *)array.data());
         if (bufRef())
         {
#ifdef MUSCLE_ENABLE_ZLIB_ENCODING
            ZLibCodec inflater;
            ByteBufferRef decompressedRef = inflater.Inflate(*bufRef());
            if (decompressedRef()) bufRef = decompressedRef;
#endif
            if (bufRef()) return settings.Unflatten(bufRef()->GetBuffer(), bufRef()->GetNumBytes());
         }
      }
   }
   return B_ERROR;
}

status_t WriteSettingsFile(const char * prefsPath, const Message & settings, int compressionLevel)
{
   ByteBufferRef bufRef = GetByteBufferFromPool(settings.FlattenedSize());
   if (bufRef())
   {
      settings.Flatten(bufRef()->GetBuffer());
      if (compressionLevel > 0)
      {
#ifdef MUSCLE_ENABLE_ZLIB_ENCODING
         ZLibCodec deflater(compressionLevel);
         ByteBufferRef compressedRef = deflater.Deflate(*bufRef(), true);
         if (compressedRef()) bufRef = compressedRef;
#endif
      }
      if (bufRef())
      {
         QFile file(ToQ(prefsPath));
         if ((file.open(IO_WriteOnly))&&(file.writeBlock((const char *)bufRef()->GetBuffer(), bufRef()->GetNumBytes()) == (int32)bufRef()->GetNumBytes())) return B_NO_ERROR;
      }
   }
   return B_ERROR;
}

QDir GetSettingsFolder(const QString & subFolderName, bool createIfNecessary)
{
   const char * SETTINGS_FOLDER_NAME = ".audiomove_settings";

   QDir dir = QDir::home();
   if (dir.exists())
   {
      if (createIfNecessary) (void) dir.mkdir(ToQ(SETTINGS_FOLDER_NAME));
      dir = QDir(dir.absPath() + ToQ("/") + ToQ(SETTINGS_FOLDER_NAME));

      if (subFolderName.length() > 0)
      {
         if (createIfNecessary) (void) dir.mkdir(subFolderName);
         dir = QDir(dir.absPath() + ToQ("/") + subFolderName);
      }
   }
   return dir;
}

static const char * REGISTRY_FILE_PREFIX = "key_";
static const char * REGISTRY_DIR_NAME    = "LCS_Registry";

status_t SaveMessageToRegistry(const char * subkey, const Message & msg)
{
   SharedMemory memLock;  // used as a read/write lock to ensure the file isn't corrupted
   String temp(REGISTRY_FILE_PREFIX); temp += subkey;
   if ((memLock.SetArea(temp(), 1) == B_NO_ERROR)&&(memLock.LockAreaReadWrite() == B_NO_ERROR))
   {
      QDir dir = GetSettingsFolder(ToQ(REGISTRY_DIR_NAME), true);
      if (dir.exists())
      {
         String path = FromQ(dir.absPath()); path += "/"; path += temp;
         return WriteSettingsFile(path(), msg, 0);  // don't worry, SharedMemory dtor unlocks the area for us
      }
   }
   return B_ERROR;
}

status_t LoadMessageFromRegistry(const char * subkey, Message & msg)
{
   SharedMemory memLock;  // used as a read/write lock to ensure the file isn't corrupted
   String temp(REGISTRY_FILE_PREFIX); temp += subkey;
   if ((memLock.SetArea(temp(), 1) == B_NO_ERROR)&&(memLock.LockAreaReadOnly() == B_NO_ERROR))
   {
      QDir dir = GetSettingsFolder(ToQ(REGISTRY_DIR_NAME), false);
      if (dir.exists())
      {
         String path = FromQ(dir.absPath()); path += "/"; path += temp;
         return ReadSettingsFile(path(), msg);  // don't worry, SharedMemory dtor unlocks the area for us
      }
   }
   return B_ERROR;
}

void ShowDialog(QWidget * w)
{
   if (w->isHidden()) w->show();
   w->raise();
   if (w->isMinimized())
   {
      // Why is all this rigamarole necessary?  :^P
      w->hide();
      w->showNormal();
   }
}

status_t SaveWindowPositionToArchive(const QWidget * window, Message & archive)
{
   QDesktopWidget * desktop = QApplication::desktop();
   if ((desktop)&&(desktop->isVirtualDesktop() == false)&&(archive.AddInt32("screen", desktop->screenNumber((QWidget *)window)) != B_NO_ERROR)) return B_ERROR;

   QRect g = window->geometry();
   return archive.AddRect("geometry", Rect(g.left(), g.top(), g.right(), g.bottom()));
}

status_t RestoreWindowPositionFromArchive(QWidget * window, const Message & archive, bool allowResize)
{
   QDesktopWidget * desktop = QApplication::desktop();
   if (desktop->isVirtualDesktop() == false)
   {
      int32 whichScreen;
      if (archive.FindInt32("screen", &whichScreen) == B_NO_ERROR)
      {
         QWidget * newScreen = desktop->screen(whichScreen);
         if (newScreen) window->reparent(newScreen, QPoint(0,0), false);
      }
   }

   Rect r;
   if (archive.FindRect("geometry", r) == B_NO_ERROR) 
   {
      QRect qr(muscleRintf(r.left()), muscleRintf(r.top()), muscleRintf(r.Width()), muscleRintf(r.Height()));
      qr.setWidth( muscleClamp(qr.width(),  window->minimumWidth(),  window->maximumWidth()));
      qr.setHeight(muscleClamp(qr.height(), window->minimumHeight(), window->maximumHeight()));

      bool onScreen = false;
      {
         int numScreens = desktop->numScreens();
         for (int i=0; i<numScreens; i++)
         {
            // Make sure we are at least partially visible on the screen!
            QRect screenRect = desktop->screenGeometry(i);
            if (qr.intersects(screenRect))
            {
               onScreen = true;
               break;
            }
         }
      }

      // If it's not visible anywhere, move it to the center of the primary screen.
      // and the user can decide where to put it.
      if (onScreen == false) qr.moveCenter(desktop->screenGeometry().center());
      if (allowResize) window->setGeometry(qr);
                  else window->move(qr.left(), qr.top());
      return B_NO_ERROR;
   }
   return B_ERROR;
}

status_t EnsureFileFolderExists(const String & fileName)
{
#if defined(WIN32)
   String path;
#else
   String path = GetPathSepChar();
#endif

   StringTokenizer tok(fileName(), GetPathSepChar()());
   const char * next;
   while((next = tok()) != NULL)
   {
      if (tok.GetRemainderOfString())  // don't do this for the file name itself!
      {
         path += next;
#if defined(WIN32)
         DWORD attributes = GetFileAttributesA(path());
         if (attributes == INVALID_FILE_ATTRIBUTES) //||(attributes & FILE_ATTRIBUTE_DIRECTORY) ==0)
         {
            if (CreateDirectoryA(path(), 0) == FALSE) return B_ERROR;
         }
#else
         DIR * dir = opendir(path());
         if (dir) closedir(dir);
             else (void) mkdir(path(), 0777);
#endif
         path += GetPathSepChar();
      }
   }
   return B_NO_ERROR;
}

String GetPathSepChar()
{
   static const String PATH_SEP_CHAR = "/";
   return PATH_SEP_CHAR;
}

};  // end namespace audiomove
