/*
** Copyright (C) 2004 Level Control Systems <jaf@lcsaudio.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
*/

#include <fcntl.h>
#include <sys/stat.h>

#if defined(WIN32)
# include "winsupport/Win32FileHandleDataIO.h"
#else
# include "dataio/FileDescriptorDataIO.h"
#endif

#include "audiomove/RawIOThread.h"
#include "audiomove/MiscFunctions.h"

namespace audiomove {

RawIOThread :: RawIOThread(const String & readPath) : _fileName(readPath), _numWriteStreams(AUDIO_STREAMS_SOURCE), _numFrames(0), _numStreams(AUDIO_STREAMS_SOURCE)
{
   // empty
}

RawIOThread :: RawIOThread(const String & writePath, uint8 writeNumStreams) : _fileName(writePath), _numWriteStreams(writeNumStreams), _numFrames(0), _numStreams(writeNumStreams)
{
   // empty
}

RawIOThread :: ~RawIOThread()
{
   CloseFileAux();
}

status_t RawIOThread :: OpenFile()
{
   CloseFile();  // just in case
   
   if (_numWriteStreams != AUDIO_STREAMS_SOURCE) EnsureFileFolderExists(_fileName);

#if defined(WIN32)
   HANDLE fd;
   if (_numWriteStreams == AUDIO_STREAMS_SOURCE)
   {
      fd = CreateFileA(
            _fileName(),
            FILE_READ_DATA,
            FILE_SHARE_READ,
            0,
            OPEN_EXISTING,
            FILE_ATTRIBUTE_NORMAL,
            0
            );
   }
   else
   {
      fd = CreateFileA(
            _fileName(),
            FILE_READ_DATA | FILE_WRITE_DATA,
            FILE_SHARE_READ,
            0,
            CREATE_ALWAYS,
            FILE_ATTRIBUTE_NORMAL,
            0
            );
   }
#else
   int fd = open(_fileName(), (_numWriteStreams == AUDIO_STREAMS_SOURCE) ? O_RDONLY : (O_WRONLY|O_CREAT|O_TRUNC), (S_IRWXU|S_IRWXG|S_IRWXO));
#endif

#if defined(WIN32)
   if (fd != INVALID_HANDLE_VALUE)
#else
   if (fd >= 0)
#endif
   {
#if defined(WIN32)
      _dataIO.SetRef(newnothrow Win32FileHandleDataIO(fd), NULL);
#else
      _dataIO.SetRef(newnothrow FileDescriptorDataIO(fd, true), NULL);
#endif
      if (_dataIO())
      {
         static const uint32 LCS_RAW_FILE_MAGIC = (((uint32)'L')<<24)|(((uint32)'C')<<16)|(((uint32)'S')<<8);
         if (_numWriteStreams == AUDIO_STREAMS_SOURCE)
         {
            // Read mode:  Read in file header
            uint32 header;
            if (_dataIO()->Read(&header, sizeof(header)) == sizeof(header))
            {
               header = B_BENDIAN_TO_HOST_INT32(header);
               if ((header & 0xFFFFFF00) == LCS_RAW_FILE_MAGIC)
               {
                  _numStreams = header & 0xFF;

                  // Now figure out how many frames file our file contains, by seeking to the end...
                  if ((_numStreams > 0)&&(_dataIO()->Seek(0, DataIO::IO_SEEK_END) == B_NO_ERROR))
                  {
                     _numFrames = (_dataIO()->GetPosition()-sizeof(header))/(sizeof(float)*_numStreams);
                     if ((_numFrames >= 0)&&(_dataIO()->Seek(sizeof(header), DataIO::IO_SEEK_SET) == B_NO_ERROR)) return B_NO_ERROR;  // success!  Seek to start of data
                  }
               }
            }
         }
         else
         {
            // Write mode:  Write out file header
            uint32 header = LCS_RAW_FILE_MAGIC | _numWriteStreams;
            header = B_HOST_TO_BENDIAN_INT32(header);
            if (_dataIO()->Write(&header, sizeof(header)) == sizeof(header))
            {
               _numFrames = 0;  // just created, so it doesn't have any frames yet
               return B_NO_ERROR;
            }
         }
      }
#if defined(WIN32)
      else CloseHandle(fd);
#else
      else close(fd);  // in case the DataIO was never created
#endif
   }
   CloseFile();

   return B_ERROR;
}

void RawIOThread :: CloseFile()
{
   CloseFileAux();
}

void RawIOThread :: CloseFileAux()
{
   _dataIO.Reset();
}

ByteBufferRef RawIOThread :: ProcessBuffer(ByteBufferRef buf, QString & retErrStr, bool isLastBuffer)
{
   if ((_dataIO())&&(buf()))
   {
      static const float HSB_MAX_AUDIO_SAMPLE_VALUE = 2147483647.0f;  // stolen from WTRXConstants.h

      const uint32 numBytes = buf()->GetNumBytes();
      float * samples = (float *) buf()->GetBuffer();

      if ((numBytes % sizeof(float)) == 0)
      {
         if (_numWriteStreams == AUDIO_STREAMS_SOURCE)
         {
            // Read from the input file into the buffer
            int32 readBytes = _dataIO()->Read(buf()->GetBuffer(), numBytes);
            if ((readBytes >= 0)&&((readBytes % sizeof(float)) == 0))
            {
               uint32 readSamples = readBytes / sizeof(float);

               // Convert buffer from LCS/little-endian to normalized/host-endian
               for (uint32 i=0; i<readSamples; i++) samples[i] = B_LENDIAN_TO_HOST_FLOAT(samples[i])/HSB_MAX_AUDIO_SAMPLE_VALUE;
          
               buf()->SetNumBytes(readSamples*sizeof(float), false);
               if (isLastBuffer) CloseFile();  // close file now so that any disk locking is unlocked ASAP
               return buf;
            }
            else retErrStr = qApp->translate("RawIOThread", "Problem reading from raw file %1/%2").arg(readBytes).arg(numBytes);
         }
         else
         {
            const uint32 numSamples = numBytes / sizeof(float);

            // Convert buffer from normalized/host-endian to LCS/little-endian
            for (uint32 i=0; i<numSamples; i++) samples[i] = B_HOST_TO_LENDIAN_FLOAT(samples[i]*HSB_MAX_AUDIO_SAMPLE_VALUE);
           
            // Write buffer to the output file
            int32 bytesWritten = _dataIO()->Write(samples, numBytes);
            if (bytesWritten == (int32)numBytes) 
            {
               _numFrames += (numSamples/_numStreams);
               if (isLastBuffer) CloseFile();  // close file now so that the file is ready for use ASAP
               return buf;
            }
            else retErrStr = qApp->translate("RawIOThread", "Problem writing to raw file %1/%2").arg(bytesWritten).arg(numBytes);
         }
      }
      else retErrStr = qApp->translate("RawIOThread", "Bad raw chunk length %1").arg(numBytes);
   }
   if (isLastBuffer) CloseFile();  // close file now so that any disk locking is unlocked ASAP
   return ByteBufferRef();  // error!
}

};  // end namespace audiomove
